use hidapi::{HidApi, HidDevice};

use super::{TouchbarError, TouchbarEventListener};
use crate::tb::{Touchbar, TouchbarEvent, TouchbarMediaEvent, TouchbarMode};
use std::{
    sync::{Arc, Mutex},
    thread::{spawn, JoinHandle},
};

const TOUCHBAR_VENDOR_ID: u16 = 0x05ac;
const TOUCHBAR_PRODUCT_ID: u16 = 0x8600;
const TOUCHBAR_MODE_PAYLOAD: u8 = 0x02;
const TOUCHBAR_MAX_PAYLOAD_SIZE: usize = 64;

pub struct TouchbarImp {
    device: Arc<Mutex<HidDevice>>,
    mode: Mutex<TouchbarMode>,
    listeners: Arc<Mutex<Vec<Box<dyn TouchbarEventListener>>>>,
    listener_handle: Option<JoinHandle<()>>,
    listener_running: Arc<Mutex<bool>>,
}

impl TouchbarImp {
    pub fn new(hid: &HidApi) -> Result<Self, TouchbarError> {
        Self::from(hid, TouchbarMode::Special)
    }

    pub fn from(hid: &HidApi, default_mode: TouchbarMode) -> Result<Self, TouchbarError> {
        let device: HidDevice = hid.open(TOUCHBAR_VENDOR_ID, TOUCHBAR_PRODUCT_ID)?;

        let mut this = TouchbarImp {
            device: Arc::new(Mutex::new(device)),
            mode: Mutex::new(default_mode),
            listeners: Arc::new(Mutex::new(vec![])),
            listener_handle: None,
            listener_running: Arc::new(Mutex::new(false)),
        };
        this.set_mode(default_mode)?;

        Ok(this)
    }
}

impl Touchbar for TouchbarImp {
    fn set_mode(&mut self, mode: TouchbarMode) -> Result<(), TouchbarError> {
        let mut mode_guard = self.mode.lock()?;
        *mode_guard = mode;

        //TODO: Do real Error handeling here
        self.device
            .lock()
            .unwrap()
            .write(&[TOUCHBAR_MODE_PAYLOAD, mode as u8])?;

        Ok(())
    }

    fn add_listener(&mut self, listener: Box<dyn super::TouchbarEventListener>) {
        //TODO: Remove Unwrap with Real Error handeling
        self.listeners.lock().unwrap().push(listener);
    }

    fn start_listening_input(&mut self) -> Result<(), TouchbarError> {
        let mut running = self.listener_running.lock()?;
        if *running {
            return Err(TouchbarError::AlreadyListening);
        }

        *running = true;

        let device = Arc::clone(&self.device);
        let running = Arc::clone(&self.listener_running);
        let listeners = Arc::clone(&self.listeners);

        self.listener_handle = Some(spawn(move || {
            //TODO: Real Error Handeling instead of unwraps
            while *running.lock().unwrap() {
                //TODO: replace unwrap with real error handeling

                match device
                    .lock()
                    .unwrap()
                    .read(&mut [0u8; TOUCHBAR_MAX_PAYLOAD_SIZE])
                {
                    Ok(data) => {
                        //println!("data: {}", data);
                        //TODO: Parse Event
                        let event = TouchbarEvent::Media(TouchbarMediaEvent::VolumeUp);
                        //TODO: remove unwrap with real error Handeling
                        let listener_lock = listeners.lock().unwrap();
                        for listener in listener_lock.iter() {
                            listener.on_event(event);
                        }
                    }
                    Err(e) => {
                        //TODO: Real Error Handeling
                        println!("Error reading data: {:?}", e);
                        break;
                    }
                }
            }
        }));

        Ok(())
    }

    fn stop_listening_input(&mut self) -> Result<(), TouchbarError> {
        todo!()
    }
}
